AmbDecoder {
	var <wBus, <xBus, <yBus, <zBus, wComp, server;
	var w, x, y, z, isPlaying, server, <channels, addAction, target, cper;
	
	*new {arg wBus, xBus, yBus, zBus, wComp = 0.707, server;
		^super.newCopyArgs(wBus, xBus, yBus, zBus, wComp, server).initAmbDecoder;
		}
	
	initAmbDecoder {
		server = server ?? Server.default;
		isPlaying = false;
		channels = IdentityDictionary.new;
		}
		
	play {arg argAddAction = 1, argTarget = 0;
		addAction = argAddAction;
		target = argTarget;
		isPlaying.not.if({
			server.serverRunning.if({
				channels.do({arg thisChan;
					thisChan.play(addAction, target)
					});
				cper = {isPlaying = false; channels.do({arg me; me.isPlaying_(false)})};
				CmdPeriod.add(cper);
				isPlaying = true;
				}, {
				"The server must be running before you 'play' an AmbDecoder. Please boot and 
					try again".warn
				});
			}, {
			"AmbDecoder already running".warn;
			})
	}
		
	stop {
		isPlaying.if({
			isPlaying = false;
			CmdPeriod.remove(cper);
			channels.do({arg thisChan;
				thisChan.stop;
				})
			})
		}
	
	at {arg bus;
		^channels[bus]
		}
	
	wBus_ {arg newBus;
		channels.do({arg thisChan;
			thisChan.wIn_(thisChan)
			});
		}

	xBus_ {arg newBus;
		channels.do({arg thisChan;
			thisChan.xIn_(thisChan)
			});
		}

	yBus_ {arg newBus;
		channels.do({arg thisChan;
			thisChan.yIn_(thisChan)
			});
		}
	zBus_ {arg newBus;
		channels.do({arg thisChan;
			thisChan.zIn_(thisChan)
			});
		}		
	clearChannels {
		isPlaying.if({
			channels.do({arg thisChan; thisChan.stop})
			});
		channels = IdentityDictionary.new;
		}
		
	addChannel {arg outbus, az, el, dist = 0;
		var channel, distances, maxDist, deltime;
		distances = [dist];
		channels.do({arg me; distances = distances.add(me.distance)});
		((maxDist = distances.maxItem) == dist).if({
			// need to recalculate all delay times
			deltime = 0;
			channels.do({arg thisChan;
				thisChan.deltime_((maxDist - thisChan.distance) * 0.002906976744186);
				})
			}, {
			deltime = (maxDist - dist) * 0.002906976744186;
			});
		channels = channels.add(outbus ->
			channel = AmbDecoderChannel(outbus, az, el, wBus, xBus, yBus, zBus, wComp, 
				dist, deltime, server)
			);
		isPlaying.if({
			channel.play(addAction, target);
			});
		^channel;
		}

	/* auto set-up some common decoders */
	
	// args are [channel, dist] pairs. Most should work with the defaults
	// dist is in meters
	
	parseData {arg data;
		var chan, az, el, bus, dist;
		#chan, az, el = data;
		#bus, dist = chan;
		this.addChannel(bus, az, el, dist);
		}
		
	stereo {arg lOut = [0, 0], rOut = [1, 0];
		this.clearChannels;
		[[lOut, 0.33pi, 0], [rOut, -0.33pi, 1]].do({arg data;
			this.parseData(data);
			})
		}
		
	quad {arg flOut = [0, 0], frOut = [1, 0], rrOut = [2, 0], rlOut = [3, 0];
		this.clearChannels;
		[[flOut, 0.25pi, 0], [frOut, -0.25pi, 0], [rrOut, -0.75pi, 0], 
			[rlOut, 0.75pi, 0]].do({arg data;
				this.parseData(data);
				});	
		}
		
	
	// front bisects a side
	hex {arg flOut = [0, 0], frOut = [1, 0], rOut = [2, 0], rrOut = [3, 0], rlOut = [4, 0], 
			lOut = [5, 0];
		this.clearChannels;
		[[flOut, 0.1666666pi, 0], [frOut, -0.16666666pi, 0], [rOut, -0.5pi, 0], 
			[rrOut, -0.83333333pi, 0], [rlOut, 0.833333333pi, 0], [lOut, 0.5pi, 0]].do({arg data;
				this.parseData(data)
				});
		}

	// front is a speaker
	hexB {arg fOut = [0, 0], frOut = [1, 0], rrOut = [2, 0], rOut = [3, 0], rlOut = [4, 0], 
			flOut = [5, 0];
		this.clearChannels;
		[[fOut, 0, 0], [frOut, -0.33333333pi, 0], [rrOut, -0.66666667pi, 0], [rOut, -pi, 0],
				[rlOut, 0.66666667pi, 0], [flOut, 0.33333333pi, 0]].do({arg data;
					this.parseData(data)
					});
		}
		
	cube {arg bflOut = [0, 0], bfrOut = [1, 0], brrOut = [2, 0], brlOut = [3, 0], tflOut = [4, 0],
			tfrOut = [5, 0], trrOut = [6, 0], trlOut = [7, 0];
		[[bflOut, 0.25pi, -0.25pi], [bfrOut, -0.25pi, -0.25pi], [brrOut, -0.75pi, -0.25pi],
			[brlOut, 0.75pi, -0.25pi], [tflOut, 0.25pi, 0.25pi], [tfrOut, -0.25pi, 0.25pi],
			[trrOut, -0.75pi, 0.25pi], [trlOut, 0.75pi, 0.25pi]].do({arg data;
				this.parseData(data)
				})
		}
	}

AmbDecoderGUI {
	var decoder, bounds, <window, <inputs, inputView, <outputs, outputView, addOutButton;
	var server, busses, width, height, spec, azspec, elspec, <offset, <decoders, numInputs;
	var <addAction, target, isPlaying, playButton;
	
	*new {arg decoder, bounds, server;
		^super.new.initAmbDecoderGUI(decoder.asArray, bounds, server);
		}
	
	*fromFile {arg path, server;
		var dict, amb, ins, outs;
		// do the whole server setup thing again
		dict = Dictionary.readArchive(path);
		amb = this.new(nil, dict[\bounds], server);
		ins = [];
		outs = [];
		dict[\inputs].do({arg thisInput; ins = ins.add(thisInput)});
		ins.sort({arg a, b; a[\slot] < b[\slot]});
		ins.do({arg thisInput;
			amb.addInput(thisInput[\w], thisInput[\x], thisInput[\y], thisInput[\z],
				thisInput[\wComp], thisInput[\scale])
				});
		dict[\outputs].do({arg thisOutput; outs = outs.add(thisOutput)});
		outs.sort({arg a, b; a[\slot] < b[\slot]});
		outs.do({arg thisOutput;
			amb.addOutput(thisOutput[\out], thisOutput[\az], thisOutput[\el], thisOutput[\dist])
			});
		amb.addAction.value_(dict[\addAction]);
		^amb;
		}
		
	save {arg path;
		var dict;
		dict = this.sendToDict;
		path.isNil.if({
			Dialog.savePanel({arg thisPath;
				dict.writeTextArchive(thisPath);
				}, {
				"No file saved".warn;
				})
			}, {
			dict.writeTextArchive(path);
			})
		}
		
	sendToDict {
		var dict, w, x, y, z, wComp, scale, out, az, el, dist;
		dict = Dictionary.new;
		dict.add(\offset -> offset.value);
		dict.add(\bounds -> bounds);
		dict.add(\inputs -> Dictionary.new);
		inputs.do({arg me, i;
			#w, x, y, z, wComp, scale = me;
			dict[\inputs].add(i -> Dictionary.new);
			dict[\inputs][i].add(\slot -> i);
			dict[\inputs][i].add(\w -> w.value);
			dict[\inputs][i].add(\x -> x.value);
			dict[\inputs][i].add(\y -> y.value);
			dict[\inputs][i].add(\z -> z.value);
			dict[\inputs][i].add(\wComp -> wComp.value);
			dict[\inputs][i].add(\scale -> scale.value);
			});
		dict.add(\outputs -> Dictionary.new);	
		/* outVal = 0, azVal = 0, elVal = 0, distanceVal = 0;*/
		outputs.do({arg me, i;
			#out, az, el, dist = me;
			dict[\outputs].add(i ->	 Dictionary.new);
			dict[\outputs][i].add(\slot -> i);
			dict[\outputs][i].add(\out -> out.value);
			dict[\outputs][i].add(\az -> az.value);
			dict[\outputs][i].add(\el -> el.value);
			dict[\outputs][i].add(\dist -> dist.value);
			});
		dict.add(\addAction -> addAction.value);
		^dict
		}
			
	initAmbDecoderGUI {arg argDecoders, argBounds, argServer;
		decoder = argDecoders; // probably just one, but needs to be an array!
		bounds = argBounds ?? {Rect(500, 450, 800, 400)};
		width = bounds.width;
		height = bounds.height;
		window = Window.new("Ambisonic Decoder", bounds).front;
		window.onClose_({isPlaying.if({this.stop})});
		window.view.decorator_(FlowLayout(Rect(0, 0, width, height)));
		server = argServer ?? {Server.default};
		busses = server.options.numAudioBusChannels.collect({arg me; me.asString});
		this.initInputPane;
		window.view.decorator.nextLine;
		this.initOutputPane;
		spec = [-inf, 12, \db].asSpec;
		azspec = [1.0, -1.0].asSpec;
		elspec = [-1.0, 1.0].asSpec;
		numInputs = 0;
		isPlaying = false;
		}
	
		
	wBus_ {arg bus, decoder = 0, changeAll = true;
		decoders.notNil.if({
			decoders[decoder].channels.do({arg thisChan; thisChan.wIn_(bus)});
			});
		inputs[decoder][0].value_(bus);
		changeAll.if({
			this.xBus_(bus + 1, decoder);
			this.yBus_(bus + 2, decoder);
			this.zBus_(bus + 3, decoder);						});
				
		}

	xBus_ {arg bus, decoder = 0;
		decoders.notNil.if({
			decoders[decoder].channels.do({arg thisChan; thisChan.xIn_(bus)});
			});
		inputs[decoder][1].value_(bus);
		}
	
	yBus_ {arg bus, decoder = 0;
		decoders.notNil.if({
			decoders[decoder].channels.do({arg thisChan; thisChan.yIn_(bus)});
			});

		inputs[decoder][2].value_(bus);
		}
	
	zBus_ {arg bus, decoder = 0;
		decoders.notNil.if({
			decoders[decoder].channels.do({arg thisChan; thisChan.zIn_(bus)});
			});
		inputs[decoder][3].value_(bus);
		}
					
	initInputPane {
		StaticText(window, (width * 0.15) @ (height * 0.05))
			.string_("Ambisonic Inputs");
		addOutButton = Button(window, (width * 0.15) @ (height * 0.05))
			.states_([
				["Add Input", Color.black, Color.white]
				])
			.action_({
				this.addInput; 
				})
			.canFocus_(false);
		playButton = Button(window, (width * 0.1) @ (height * 0.05))
			.states_([
				[">", Color.black, Color.white],
				["[]", Color.white, Color.black]
				])
			.action_({arg me;
				this.playGUI(me.value);
				});
		StaticText(window, (width * 0.1) @ (height * 0.05))
			.string_(server.name); 
		addAction = PopUpMenu(window, (width * 0.1) @ (height * 0.05))
			.items_(["head", "tail", "before", "after", "replace"])
			.value_(3);
		StaticText(window, (width * 0.1) @ (height * 0.05))
			.string_("Target:");
		target = NumberBox(window, (width * 0.1) @ (height * 0.05))
			.value_(1); 
		Button(window, (width * 0.1) @ (height * 0.05))
			.states_([
				["Save Decoder", Color.white, Color.black]
				])
			.action_({
				this.save
				});
		window.view.decorator.nextLine;
		inputView = ScrollView(window, (width * 0.99) @ (height * 0.3));
		inputView.decorator_(FlowLayout(Rect(0, 0, inputView.bounds.width, 
			inputView.bounds.height)));
		inputView.background_(Color.new255(139, 62, 47));
		}
		
	addInput {arg wBus = 0, xBus = 0, yBus = 0, zBus = 0, wCompVal = 0.707, scaleVal = 0.0;
		var w, x, y, z, wComp, scaleSlider, wCompSlider, scale, tmp;
		tmp = numInputs;
		StaticText(inputView, 20 @ 20).string_("W").stringColor_(Color.white);
		w = PopUpMenu(inputView, 40 @ 20)
			.items_(busses)
			.stringColor_(Color.white)
			.value_(wBus)
			.action_({arg me;
				x.value_(me.value + 1);
				y.value_(me.value + 2);
				z.value_(me.value + 3);
				});
		StaticText(inputView, 20 @ 20).string_("X").stringColor_(Color.white);
		x = PopUpMenu(inputView, 40 @ 20)
			.items_(busses)
			.stringColor_(Color.white)
			.value_(xBus);
		StaticText(inputView, 20 @ 20).string_("Y").stringColor_(Color.white);
		y = PopUpMenu(inputView, 40 @ 20)
			.items_(busses)
			.stringColor_(Color.white)
			.value_(yBus);
		StaticText(inputView, 20 @ 20).string_("Z").stringColor_(Color.white);
		z = PopUpMenu(inputView, 40 @ 20)
			.items_(busses)
			.stringColor_(Color.white)
			.value_(zBus);
		StaticText(inputView, 60 @ 20).string_("W comp").stringColor_(Color.white);
		wCompSlider = Slider(inputView, 100 @ 20)
			.value_(wCompVal)
			.action_({arg me; wComp.valueAction_(me.value)});
		wComp = NumberBox(inputView, 40 @ 20)
			.value_(wCompVal)
			.action_({arg me; wCompSlider.value_(me.value);
				decoders[tmp].channels.do({arg thisChannel;
					thisChannel.wComp_(me.value)
					})
				});
		StaticText(inputView, 120 @ 20).string_("Amp scale (in db)").stringColor_(Color.white);
		scaleSlider = Slider(inputView, 100 @ 20)
			.value_(spec.unmap(scaleVal))
			.action_({arg me; scale.valueAction_(spec.map(me.value))});
		scale = NumberBox(inputView, 40 @ 20)
			.value_(scaleVal)
			.action_({arg me; scaleSlider.value_(spec.unmap(me.value));
				decoders[tmp].channels.do({arg thisChannel;
					thisChannel.amp_(me.value.dbamp)
					})
				});
		inputs = inputs.add([w, x, y, z, wComp, scale]);
		numInputs = numInputs + 1;
		inputView.decorator.nextLine;
		}
		
	initOutputPane {
		StaticText(window, (width * 0.15) @ (height * 0.05))
			.string_("Speaker Outputs");
		addOutButton = Button(window, (width * 0.15) @ (height * 0.05))
			.states_([
				["Add Speaker", Color.black, Color.white]
				])
			.action_({
				this.addOutput; 
				})
			.canFocus_(false);
		Button.new(window, (width * 0.1) @ (height * 0.05))
			.states_([["Stereo", Color.black, Color.white]])
			.action_({
				this.addOutput(0, 0.33, 0, 0);
				this.addOutput(1, -0.33, 0, 0);	
				});
		Button.new(window, (width * 0.1) @ (height * 0.05))
			.states_([["Quad", Color.black, Color.white]])
			.action_({
				this.addOutput(0, 0.25, 0, 0);
				this.addOutput(1, -0.25, 0, 0);	
				this.addOutput(2, -0.75, 0, 0);	
				this.addOutput(3, 0.75, 0, 0);	
				});
		Button.new(window, (width * 0.1) @ (height * 0.05))
			.states_([["Hex", Color.black, Color.white]])
			.action_({
				this.addOutput(0, 0.1666667, 0, 0);
				this.addOutput(1, -0.1666667, 0, 0);	
				this.addOutput(2, -0.5, 0, 0);	
				this.addOutput(3, -0.8333333, 0, 0);	
				this.addOutput(4, 0.833333, 0, 0);
				this.addOutput(5, 0.5, 0, 0);	
				});
		Button.new(window, (width * 0.1) @ (height * 0.05))
			.states_([["HexB", Color.black, Color.white]])
			.action_({
				this.addOutput(0, 0, 0, 0);
				this.addOutput(1, -0.333333, 0, 0);	
				this.addOutput(2, -0.666667, 0, 0);	
				this.addOutput(3, -1, 0, 0);	
				this.addOutput(4, 0.666667, 0, 0);
				this.addOutput(5, 0.333333, 0, 0);	
				});
		Button.new(window, (width * 0.1) @ (height * 0.05))
			.states_([["Cube", Color.black, Color.white]])
			.action_({
				this.addOutput(0, 0.25, -0.25, 0);
				this.addOutput(1, -0.25, -0.25, 0);	
				this.addOutput(2, -0.75, -0.25, 0);	
				this.addOutput(3, 0.75, -0.25, 0);	
				this.addOutput(4, 0.25, 0.25, 0);
				this.addOutput(5, -0.25, 0.25, 0);	
				this.addOutput(6, -0.75, 0.25, 0);	
				this.addOutput(7, 0.75, 0.25, 0);					});
		Button.new(window, (width * 0.1) @ (height * 0.05))
			.states_([["Double Hex", Color.black, Color.white]])
			.action_({
				this.addOutput(0, 0.1666667, -0.25, 0);
				this.addOutput(1, -0.1666667, -0.25, 0);
				this.addOutput(2, -0.5, -0.25, 0);	
				this.addOutput(3, -0.8333333, -0.25, 0);
				this.addOutput(4, 0.833333, -0.25, 0);
				this.addOutput(5, 0.5, -0.25, 0);
				this.addOutput(6, 0.1666667, 0.25, 0);
				this.addOutput(7, -0.1666667, 0.25, 0);
				this.addOutput(8, -0.5, 0.25, 0);	
				this.addOutput(9, -0.8333333, 0.25, 0);
				this.addOutput(10, 0.833333, 0.25, 0);
				this.addOutput(11, 0.5, 0.25, 0);				});
		window.view.decorator.nextLine;
		StaticText(window, (width * 0.15) @ (height * 0.05))
			.string_("Speaker Offset");
		offset = NumberBox(window, (width * 0.1) @ (height * 0.05))
			.value_(0)
			.action_({arg num;
				outputs.do({arg me, i;
					me[0].value_(num.value + i)
					})
				});
		window.view.decorator.nextLine;
		outputView = ScrollView(window, (width * 0.99) @ (height  * 0.49));
		outputView.decorator_(FlowLayout(Rect(0, 0, outputView.bounds.width,
			outputView.bounds.height)));
		outputView.background_(Color.new255(85, 107, 47))
		}
		
	addOutput {arg outVal = 0, azVal = 0, elVal = 0, distanceVal = 0;
		var outbus, azSlider, az, elSlider, el, distance;
		StaticText(outputView, 60 @ 20).string_("Outbus").stringColor_(Color.black);
		outbus = PopUpMenu(outputView, 40 @ 20)
			.items_(busses)
			.value_(outVal + offset.value)
			.stringColor_(Color.black);
		StaticText(outputView, 80 @ 20).string_("Azimuth (* pi)").stringColor_(Color.black);
		azSlider = Slider(outputView, 100 @ 20)
			.value_(azspec.unmap(azVal))
			.action_({arg me; az.value_(azspec.map(me.value))});
		az = NumberBox(outputView, 40 @ 20)
			.value_(azVal)
			.action_({arg me; azSlider.value_(azspec.unmap(me.value))});
		StaticText(outputView, 80 @ 20).string_("Elevation (* pi)").stringColor_(Color.black);
		elSlider = Slider(outputView, 120 @ 20)
			.value_(elspec.unmap(elVal))
			.action_({arg me; el.value_(elspec.map(me.value))});
		el = NumberBox(outputView, 40 @ 20)
			.value_(elVal)
			.action_({arg me; elSlider.value_(elspec.unmap(me.value))});
		StaticText(outputView, 120 @ 20).string_("Distance in meters").stringColor_(Color.black);
		distance = NumberBox(outputView, 40 @ 20)
			.value_(distanceVal);
		outputView.decorator.nextLine;
		outputs = outputs.add([outbus, az, el, distance]);
		isPlaying.if({
			decoders.do({arg thisDecoder;
				thisDecoder.addChannel(outbus.value, az.value * pi, el.value * pi, distance.value);
				})
			});
		}
		
	playGUI {arg value = 1;
		(value == 0).if({
			isPlaying = false;
			decoders.do({arg me; me.stop});
			playButton.value_(0);
			}, {
			playButton.value_(1);
			isPlaying = true;
			decoders = [];
			inputs.do({arg me;
				decoders = decoders.add(
					AmbDecoder(me[0].value, me[1].value, me[2].value, me[3].value, 
						(me[4].value == 1).if({1.0}, {0.707}), server))
					});
			outputs.do({arg me;
				decoders.do({arg thisDecoder;
					thisDecoder.addChannel(me[0].value, me[1].value * pi, me[2].value * pi,
						me[3].value);
						})
					});
			decoders.do({arg me; me.play(addAction.value, target.value.asInteger)});
			})	
		}
		
	stop {
		decoders.do({arg me; me.stop});
		}
	}
	
/* Really - don't try to make these yourself... they are for the above classes */

AmbDecoderChannel {
	var outbus, <az, <el, wIn, xIn, yIn, zIn, wComp, <distance, <deltime, server;
	var nodeID, <>isPlaying;
	classvar sd;
	
	*new {arg outbus, az = 0, el = 0, wIn, xIn, yIn, zIn, wComp = 0.707, distance, 
			deltime = 0, server;
		^super.newCopyArgs(outbus, az, el, wIn, xIn, yIn, zIn, wComp, distance, deltime, server)
			.initAmbDecoderChannel;
		}
		
	initAmbDecoderChannel {
		server = server ?? Server.default;
		sd.load(server);
		isPlaying = false;
		}

	play {arg addAction = 1, target = 0;
		isPlaying = true;
		server.sendMsg(\s_new, \ambDecoderChannelSD, nodeID = server.nextNodeID,
			addAction, target, \outbus, outbus, \az, az, \el, el, \wIn, wIn,
			\xIn, xIn, \yIn, yIn, \zIn, zIn, \wComp, wComp, \deltime, deltime);
		}
		
	stop {
		isPlaying = false;
		server.sendMsg(\n_set, nodeID, \thisGate, 0)
		}
		
	deltime_ {arg newDeltime;
		deltime = newDeltime;
		isPlaying.if({
			server.sendMsg(\n_set, nodeID, \deltime, deltime)
			});
		}
		
	az_ {arg newAz;
		az = newAz;
		isPlaying.if({
			server.sendMsg(\n_set, nodeID, \az, az)
			})
		}
		
	el_ {arg newEl;
		el = newEl;
		isPlaying.if({
			server.sendMsg(\n_set, nodeID, \el, el)
			})
		}
		
	amp_ {arg newAmp;
		isPlaying.if({
			server.sendMsg(\n_set, nodeID, \amp, newAmp)
			})
		}
	
	wIn_ {arg newW;
		wIn = newW;
		isPlaying.if({
			server.sendMsg(\n_set, nodeID, \wIn, wIn)
			});
		}

	xIn_ {arg newX;
		xIn = newX;
		isPlaying.if({
			server.sendMsg(\n_set, nodeID, \xIn, xIn)
			});
		}
		
	yIn_ {arg newY;
		yIn = newY;
		isPlaying.if({
			server.sendMsg(\n_set, nodeID, \yIn, yIn)
			});
		}
		
	zIn_ {arg newZ;
		zIn = newZ;
		isPlaying.if({
			server.sendMsg(\n_set, nodeID, \zIn, zIn)
			});
		}
		
	wComp_ {arg newWComp;
		wComp = newWComp;
		isPlaying.if({
			server.sendMsg(\n_set, nodeID, \wComp, wComp)
			})
		}
				
	*initClass {
		sd = SynthDef(\ambDecoderChannelSD, {arg thisGate = 1, outbus, az, el, wIn, xIn, yIn, 
					zIn, wComp = 0.707, amp = 1, deltime = 0;
				var env, dec, w, x, y, z, del;
				w = In.ar(wIn, 1) * Lag.kr(wComp, 0.1);
				x = In.ar(xIn, 1);
				y = In.ar(yIn, 1);
				z = In.ar(zIn, 1);
				env = EnvGen.kr(Env([0, 1, 0], [0.1, 0.1], \lin, 1), thisGate, doneAction: 2);
				// compensate for W above... allows variable compensation values
				dec = BFDecode1.ar(w, x, y, z, az, el, 1) * Lag.kr(amp, 0.1) * env; 
				del = DelayN.ar(dec, deltime, deltime) * env; // yes, I mean to mul the env twice
				Out.ar(outbus, dec);
			})
		}
			
	}